"""Fixed key modelset index

Peek Plugin Database Migration Script

Revision ID: ab9da4532175
Revises: b0b08aff8542
Create Date: 2017-07-16 13:38:20.517763

"""

# revision identifiers, used by Alembic.
revision = 'ab9da4532175'
down_revision = 'b0b08aff8542'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import geoalchemy2


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_index('idx_LiveDbDKey_modelSet_key', 'LiveDbItem', ['modelSetId', 'key'], unique=True, schema='pl_livedb')
    op.drop_index('idx_LiveDbDKey_liveDbKey', table_name='LiveDbItem', schema='pl_livedb')
    op.drop_index('idx_LiveDbDKey_modelSetId', table_name='LiveDbItem', schema='pl_livedb')
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_index('idx_LiveDbDKey_modelSetId', 'LiveDbItem', ['modelSetId'], unique=False, schema='pl_livedb')
    op.create_index('idx_LiveDbDKey_liveDbKey', 'LiveDbItem', ['key'], unique=False, schema='pl_livedb')
    op.drop_index('idx_LiveDbDKey_modelSet_key', table_name='LiveDbItem', schema='pl_livedb')
    # ### end Alembic commands ###



from peek_plugin_livedb._private.PluginNames import livedbTuplePrefix
from sqlalchemy import Column
from sqlalchemy import Integer, String
from vortex.Tuple import addTupleType, Tuple, TupleField

from .DeclarativeBase import DeclarativeBase


@addTupleType
class LiveDbModelSet(Tuple, DeclarativeBase):
    __tablename__ = 'LiveDbModelSet'
    __tupleType__ = livedbTuplePrefix + __tablename__

    id = Column(Integer, primary_key=True, autoincrement=True)
    name = Column(String, nullable=False, unique=True)
    comment = Column(String)

    propsJson = Column(String)


def getOrCreateLiveDbModelSet(session, modelSetName:str) -> LiveDbModelSet:
    qry = session.query(LiveDbModelSet).filter(LiveDbModelSet.name == modelSetName)
    if not qry.count():
        session.add(LiveDbModelSet(name=modelSetName))
        session.commit()

    return qry.one()



"""Added unique constraing to model set

Peek Plugin Database Migration Script

Revision ID: b0b08aff8542
Revises: 64309a77b48d
Create Date: 2017-07-13 20:48:40.825705

"""

# revision identifiers, used by Alembic.
revision = 'b0b08aff8542'
down_revision = '64309a77b48d'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import geoalchemy2


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_unique_constraint(None, 'LiveDbModelSet', ['name'], schema='pl_livedb')
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_constraint(None, 'LiveDbModelSet', schema='pl_livedb', type_='unique')
    # ### end Alembic commands ###


"""Removed string lengths

Peek Plugin Database Migration Script

Revision ID: 19b2a8d326ad
Revises: ab9da4532175
Create Date: 2019-06-10 21:52:38.648849

"""

# revision identifiers, used by Alembic.
revision = '19b2a8d326ad'
down_revision = 'ab9da4532175'
branch_labels = None
depends_on = None

import sqlalchemy as sa
from alembic import op


def upgrade():
    op.alter_column('LiveDbItem', 'key', type_=sa.String(), nullable=False,
                    schema='pl_livedb')
    op.alter_column('LiveDbItem', 'rawValue', type_=sa.String(), nullable=True,
                    schema='pl_livedb')
    op.alter_column('LiveDbItem', 'displayValue', type_=sa.String(), nullable=True,
                    schema='pl_livedb')
    op.alter_column('LiveDbItem', 'importHash', type_=sa.String(), nullable=True,
                    schema='pl_livedb')
    op.alter_column('LiveDbItem', 'propsJson', type_=sa.String(), nullable=True,
                    schema='pl_livedb')

    op.alter_column('LiveDbModelSet', 'name', type_=sa.String(), nullable=False,
                    schema='pl_livedb')
    op.alter_column('LiveDbModelSet', 'propsJson', type_=sa.String(), nullable=True,
                    schema='pl_livedb')


def downgrade():
    pass



from peek_plugin_livedb._private.PluginNames import livedbTuplePrefix

from vortex.Tuple import Tuple, addTupleType


@addTupleType
class LiveDbRawValueUpdateTuple(Tuple):
    """ Live DB Raw Value Update

    This tuple represents an update to the raw value for a live db item

    """
    __tupleType__ = livedbTuplePrefix + 'LiveDbRawValueUpdateTuple'
    __slots__ = ("key", "rawValue")



import logging

from sqlalchemy import Column, text
from sqlalchemy import ForeignKey
from sqlalchemy import Integer, String
from sqlalchemy.orm import relationship
from sqlalchemy.sql.schema import Index, Sequence

from peek_plugin_livedb._private.PluginNames import livedbTuplePrefix
from vortex.Tuple import Tuple, addTupleType, JSON_EXCLUDE
from .DeclarativeBase import DeclarativeBase
from .LiveDbModelSet import LiveDbModelSet

logger = logging.getLogger(__name__)


@addTupleType
class LiveDbItem(Tuple, DeclarativeBase):
    __tupleTypeShort__ = 'LDK'
    __tablename__ = 'LiveDbItem'
    __tupleType__ = livedbTuplePrefix + __tablename__

    NUMBER_VALUE = 0
    STRING_VALUE = 1
    COLOR = 2
    LINE_WIDTH = 3
    LINE_STYLE = 4
    GROUP_PTR = 5

    id_seq = Sequence('LiveDbItem_id_seq',
                      metadata=DeclarativeBase.metadata,
                      schema=DeclarativeBase.metadata.schema)
    id = Column(Integer, id_seq, server_default=id_seq.next_value(),
                primary_key=True, autoincrement=False)

    modelSetId = Column(Integer, ForeignKey('LiveDbModelSet.id', ondelete='CASCADE'),
                        doc=JSON_EXCLUDE, nullable=False)
    modelSet = relationship(LiveDbModelSet)

    # comment="The unique reference of the value we want from the live db"
    key = Column(String, nullable=False)

    # comment="The last value from the source"
    rawValue = Column(String)

    # comment="The PEEK value, converted to PEEK IDs if required (Color for example)"
    displayValue = Column(String)

    # comment="The type of data this value represents"
    dataType = Column(Integer, nullable=False)

    importHash = Column(String)

    # Store custom props for this link
    propsJson = Column(String)

    __table_args__ = (
        Index("idx_LiveDbDKey_importHash", importHash, unique=False),
        Index("idx_LiveDbDKey_modelSet_key", modelSetId, key, unique=True),
    )




"""Initial

Peek Plugin Database Migration Script

Revision ID: 64309a77b48d
Revises: 
Create Date: 2017-07-13 16:22:26.973631

"""

# revision identifiers, used by Alembic.
revision = '64309a77b48d'
down_revision = None
branch_labels = None
depends_on = None


from sqlalchemy.dialects.mssql.base import MSDialect
from sqlalchemy.dialects.postgresql.base import PGDialect
from sqlalchemy.schema import Sequence, CreateSequence
from alembic import op
import sqlalchemy as sa
import geoalchemy2


def isMssqlDialect():
    return isinstance(op.get_bind().engine.dialect, MSDialect)

def isPostGreSQLDialect():
    return isinstance(op.get_bind().engine.dialect, PGDialect)

def nextval(seqName):
    if isMssqlDialect():
        return sa.text('(NEXT VALUE FOR "pl_livedb"."%s")' % seqName)

    return sa.text('nextval(\'pl_livedb."%s"\')' % seqName)

def upgrade():
    op.execute(CreateSequence(Sequence('LiveDbItem_id_seq',  minvalue=0, schema='pl_livedb')))

    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table('LiveDbModelSet',
    sa.Column('id', sa.Integer(), nullable=False),
    sa.Column('name', sa.String(length=50), nullable=False),
    sa.Column('comment', sa.String(), nullable=True),
    sa.Column('propsJson', sa.String(length=500), nullable=True),
    sa.PrimaryKeyConstraint('id'),
    schema='pl_livedb'
    )
    op.create_table('Setting',
    sa.Column('id', sa.Integer(), nullable=False),
    sa.Column('name', sa.String(length=50), nullable=True),
    sa.PrimaryKeyConstraint('id'),
    schema='pl_livedb'
    )



    op.create_table('LiveDbItem',
    sa.Column('id', sa.Integer(), server_default=nextval("LiveDbItem_id_seq"), nullable=False),
    sa.Column('modelSetId', sa.Integer(), nullable=False),
    sa.Column('key', sa.String(length=50), nullable=False),
    sa.Column('rawValue', sa.String(length=255), nullable=True),
    sa.Column('displayValue', sa.String(length=255), nullable=True),
    sa.Column('dataType', sa.Integer(), nullable=False),
    sa.Column('importHash', sa.String(length=100), nullable=True),
    sa.Column('propsJson', sa.String(length=500), nullable=True),
    sa.ForeignKeyConstraint(['modelSetId'], ['pl_livedb.LiveDbModelSet.id'], ondelete='CASCADE'),
    sa.PrimaryKeyConstraint('id'),
    schema='pl_livedb'
    )
    op.create_index('idx_LiveDbDKey_importHash', 'LiveDbItem', ['importHash'], unique=False, schema='pl_livedb')
    op.create_index('idx_LiveDbDKey_liveDbKey', 'LiveDbItem', ['key'], unique=False, schema='pl_livedb')
    op.create_index('idx_LiveDbDKey_modelSetId', 'LiveDbItem', ['modelSetId'], unique=False, schema='pl_livedb')
    op.create_table('SettingProperty',
    sa.Column('id', sa.Integer(), nullable=False),
    sa.Column('settingId', sa.Integer(), nullable=False),
    sa.Column('key', sa.String(length=50), nullable=False),
    sa.Column('type', sa.String(length=16), nullable=True),
    sa.Column('int_value', sa.Integer(), nullable=True),
    sa.Column('char_value', sa.String(length=50), nullable=True),
    sa.Column('boolean_value', sa.Boolean(), nullable=True),
    sa.ForeignKeyConstraint(['settingId'], ['pl_livedb.Setting.id'], ),
    sa.PrimaryKeyConstraint('id'),
    schema='pl_livedb'
    )
    op.create_index('idx_SettingProperty_settingId', 'SettingProperty', ['settingId'], unique=False, schema='pl_livedb')
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_index('idx_SettingProperty_settingId', table_name='SettingProperty', schema='pl_livedb')
    op.drop_table('SettingProperty', schema='pl_livedb')
    op.drop_index('idx_LiveDbDKey_modelSetId', table_name='LiveDbItem', schema='pl_livedb')
    op.drop_index('idx_LiveDbDKey_liveDbKey', table_name='LiveDbItem', schema='pl_livedb')
    op.drop_index('idx_LiveDbDKey_importHash', table_name='LiveDbItem', schema='pl_livedb')
    op.drop_table('LiveDbItem', schema='pl_livedb')
    op.drop_table('Setting', schema='pl_livedb')
    op.drop_table('LiveDbModelSet', schema='pl_livedb')
    # ### end Alembic commands ###


import logging
from datetime import datetime
from typing import List

from sqlalchemy import bindparam
from sqlalchemy.sql.expression import and_
from twisted.internet import defer
from twisted.internet.defer import Deferred, inlineCallbacks

from peek_plugin_livedb._private.server.LiveDBReadApi import LiveDBReadApi
from peek_plugin_livedb._private.server.controller.LiveDbController import \
    LiveDbController
from peek_plugin_livedb._private.server.controller.LiveDbImportController import \
    LiveDbImportController
from peek_plugin_livedb._private.storage.LiveDbItem import LiveDbItem
from peek_plugin_livedb._private.storage.LiveDbModelSet import getOrCreateLiveDbModelSet
from peek_plugin_livedb._private.worker.tasks.LiveDbItemUpdateTask import updateValues
from peek_plugin_livedb.server.LiveDBWriteApiABC import LiveDBWriteApiABC
from peek_plugin_livedb.tuples.ImportLiveDbItemTuple import ImportLiveDbItemTuple
from peek_plugin_livedb.tuples.LiveDbDisplayValueUpdateTuple import \
    LiveDbDisplayValueUpdateTuple
from peek_plugin_livedb.tuples.LiveDbRawValueUpdateTuple import LiveDbRawValueUpdateTuple
from vortex.DeferUtil import deferToThreadWrapWithLogger
from vortex.Payload import Payload

logger = logging.getLogger(__name__)


class LiveDBWriteApi(LiveDBWriteApiABC):
    def __init__(self, liveDbController: LiveDbController,
                 liveDbImportController: LiveDbImportController,
                 readApi: LiveDBReadApi,
                 dbSessionCreator,
                 dbEngine):
        self._liveDbController = liveDbController
        self._liveDbImportController = liveDbImportController
        self._readApi = readApi
        self._dbSessionCreator = dbSessionCreator
        self._dbEngine = dbEngine

    def shutdown(self):
        pass

    @inlineCallbacks
    def updateRawValues(self, modelSetName: str,
                        updates: List[LiveDbRawValueUpdateTuple]) -> Deferred:
        """ Update Raw Values

        """
        if not updates:
            return

        yield updateValues.delay(modelSetName, updates, raw=True)
        self._readApi.rawValueUpdatesObservable(modelSetName).on_next(updates)

    @inlineCallbacks
    def updateDisplayValue(self, modelSetName: str,
                           updates: List[LiveDbDisplayValueUpdateTuple]) -> Deferred:
        """ Update Display Values

        """
        if not updates:
            return

        yield updateValues.delay(modelSetName, updates, raw=False)
        self._readApi.displayValueUpdatesObservable(modelSetName).on_next(updates)

    def importLiveDbItems(self, modelSetName: str,
                          newItems: List[ImportLiveDbItemTuple]) -> Deferred:
        if not newItems:
            return defer.succeed(True)

        return self._liveDbImportController.importLiveDbItems(modelSetName, newItems)

    def prioritiseLiveDbValueAcquisition(self, modelSetName: str,
                                         liveDbKeys: List[str]) -> Deferred:
        self._readApi.priorityKeysObservable(modelSetName).on_next(liveDbKeys)
        return defer.succeed(True)

    def pollLiveDbValueAcquisition(self, modelSetName: str,
                                         liveDbKeys: List[str]) -> Deferred:
        self._readApi.pollKeysObservable(modelSetName).on_next(liveDbKeys)
        return defer.succeed(True)




import logging

from peek_plugin_base.worker.PluginWorkerEntryHookABC import PluginWorkerEntryHookABC
from peek_plugin_livedb._private.storage.DeclarativeBase import loadStorageTuples
from peek_plugin_livedb._private.tuples import loadPrivateTuples
from peek_plugin_livedb._private.worker.tasks import LiveDbItemImportTask, \
    LiveDbItemUpdateTask, BulkLoadChunkTask
from peek_plugin_livedb.tuples import loadPublicTuples

logger = logging.getLogger(__name__)


class WorkerEntryHook(PluginWorkerEntryHookABC):
    def load(self):
        loadStorageTuples()
        loadPrivateTuples()
        loadPublicTuples()

        logger.debug("loaded")

    def start(self):
        logger.debug("started")

    def stop(self):
        logger.debug("stopped")

    def unload(self):
        logger.debug("unloaded")

    @property
    def celeryAppIncludes(self):
        return [LiveDbItemImportTask.__name__,
                LiveDbItemUpdateTask.__name__,
                BulkLoadChunkTask.__name__]

    @property
    def celeryApp(self):
        from .CeleryApp import celeryApp
        return celeryApp



import logging
from typing import List

from sqlalchemy import select
from txcelery.defer import DeferrableTask

from peek_plugin_base.worker import CeleryDbConn
from peek_plugin_livedb._private.storage.LiveDbItem import LiveDbItem
from peek_plugin_livedb._private.worker.CeleryApp import celeryApp
from peek_plugin_livedb.tuples.LiveDbDisplayValueTuple import LiveDbDisplayValueTuple

logger = logging.getLogger(__name__)


@DeferrableTask
@celeryApp.task(bind=True)
def qryChunkInWorker(self, offset, limit) -> List[LiveDbDisplayValueTuple]:
    """ Query Chunk

    This returns a chunk of LiveDB items from the database

    :param self: A celery reference to this task
    :param offset: The offset of the chunk
    :param limit: An encoded payload containing the updates
    :returns: List[LiveDbDisplayValueTuple] serialised in a payload json
    """

    table = LiveDbItem.__table__
    cols = [table.c.key, table.c.dataType, table.c.rawValue, table.c.displayValue]

    session = CeleryDbConn.getDbSession()
    try:
        result = session.execute(select(cols)
                                 .order_by(table.c.id)
                                 .offset(offset)
                                 .limit(limit))

        return [LiveDbDisplayValueTuple(
            key=o.key, dataType=o.dataType,
            rawValue=o.rawValue, displayValue=o.displayValue) for o in result.fetchall()]

    finally:
        session.close()



import logging
from datetime import datetime

import pytz
from collections import namedtuple
from sqlalchemy.sql.expression import bindparam, and_
from txcelery.defer import CeleryClient, DeferrableTask

from peek_plugin_base.worker import CeleryDbConn
from peek_plugin_livedb._private.storage.LiveDbItem import LiveDbItem
from peek_plugin_livedb._private.storage.LiveDbModelSet import getOrCreateLiveDbModelSet
from peek_plugin_livedb._private.worker.CeleryApp import celeryApp
from vortex.Payload import Payload

logger = logging.getLogger(__name__)


@DeferrableTask
@celeryApp.task(bind=True)
def updateValues(self, modelSetName, updates, raw=True):
    """ Compile Grids Task

    :param self: A celery reference to this task
    :param modelSetName: The model set name
    :param updates: An encoded payload containing the updates
    :param raw: Are the updates raw updates?
    :returns: A list of grid keys that have been updated.
    """

    startTime = datetime.now(pytz.utc)
    table = LiveDbItem.__table__

    session = CeleryDbConn.getDbSession()
    conn = CeleryDbConn.getDbEngine().connect()
    try:
        liveDbModelSet = getOrCreateLiveDbModelSet(session, modelSetName)

        sql = (table.update()
               .where(and_(table.c.key == bindparam('b_key'),
                           table.c.modelSetId == liveDbModelSet.id))
               .values({"rawValue" if raw else "displayValue": bindparam("b_value")}))

        conn.execute(sql, [
            dict(b_key=o.key, b_value=(o.rawValue if raw else o.displayValue))
            for o in updates])

        logger.debug("Updated %s %s values, in %s",
                     len(updates),
                     "raw" if raw else "display",
                     (datetime.now(pytz.utc) - startTime))

    except Exception as e:
        logger.exception(e)
        raise self.retry(exc=e, countdown=2)

    finally:
        session.close()
        conn.close()



from abc import ABCMeta, abstractmethod

from peek_plugin_livedb.server.LiveDBReadApiABC import LiveDBReadApiABC
from peek_plugin_livedb.server.LiveDBWriteApiABC import LiveDBWriteApiABC


class LiveDBApiABC(metaclass=ABCMeta):
    @property
    @abstractmethod
    def writeApi(self) -> LiveDBWriteApiABC:
        """ LiveDB Write API

        This API is for all the methods that make changes to the LiveDB

        :return: A reference to the LiveDBWriteApiABC instance

        """

    @property
    @abstractmethod
    def readApi(self) -> LiveDBReadApiABC:
        """ LiveDB Read API

        This API is for all the methods to read changes from the LiveDB

        :return: A reference to the LiveDBReadApiABC instance

        """



livedbFilt = {"plugin": "peek_plugin_livedb"}
livedbTuplePrefix = "peek_plugin_livedb."
livedbObservableName = "peek_plugin_livedb"
livedbActionProcessorName = "peek_plugin_livedb"
livedbTupleOfflineServiceName = "peek_plugin_livedb"


from peek_plugin_livedb._private.PluginNames import livedbFilt
from peek_plugin_livedb._private.PluginNames import livedbObservableName
from vortex.handler.TupleDataObservableHandler import TupleDataObservableHandler


def makeTupleDataObservableHandler(ormSessionCreator):
    """" Make Tuple Data Observable Handler

    This method creates the observable object, registers the tuple providers and then
    returns it.

    :param ormSessionCreator: A function that returns a SQLAlchemy session when called

    :return: An instance of :code:`TupleDataObservableHandler`

    """
    tupleObservable = TupleDataObservableHandler(
        observableName=livedbObservableName,
        additionalFilt=livedbFilt)

    # # Register TupleProviders here
    # tupleObservable.addTupleProvider(StringIntTuple.tupleName(),
    #                                  StringIntTupleProvider(ormSessionCreator))
    return tupleObservable



import logging

from celery import Celery

from peek_plugin_base.server.PluginServerEntryHookABC import PluginServerEntryHookABC
from peek_plugin_base.server.PluginServerStorageEntryHookABC import \
    PluginServerStorageEntryHookABC
from peek_plugin_base.server.PluginServerWorkerEntryHookABC import \
    PluginServerWorkerEntryHookABC
from peek_plugin_livedb._private.server.controller.LiveDbController import \
    LiveDbController
from peek_plugin_livedb._private.server.controller.LiveDbImportController import \
    LiveDbImportController
from peek_plugin_livedb._private.storage import DeclarativeBase
from peek_plugin_livedb._private.storage.DeclarativeBase import loadStorageTuples
from peek_plugin_livedb._private.tuples import loadPrivateTuples
from peek_plugin_livedb.tuples import loadPublicTuples
from .LiveDBApi import LiveDBApi
from .TupleActionProcessor import makeTupleActionProcessorHandler
from .TupleDataObservable import makeTupleDataObservableHandler
from .admin_backend import makeAdminBackendHandlers
from .controller.MainController import MainController

logger = logging.getLogger(__name__)


class ServerEntryHook(PluginServerEntryHookABC, PluginServerStorageEntryHookABC,
                      PluginServerWorkerEntryHookABC):
    def __init__(self, *args, **kwargs):
        """" Constructor """
        # Call the base classes constructor
        PluginServerEntryHookABC.__init__(self, *args, **kwargs)

        #: Loaded Objects, This is a list of all objects created when we start
        self._loadedObjects = []

        self._api = None

    def load(self) -> None:
        """ Load

        This will be called when the plugin is loaded, just after the db is migrated.
        Place any custom initialiastion steps here.

        """
        loadStorageTuples()
        loadPrivateTuples()
        loadPublicTuples()
        logger.debug("Loaded")

    @property
    def dbMetadata(self):
        return DeclarativeBase.metadata

    def start(self):
        """ Start

        This will be called when the plugin is loaded, just after the db is migrated.
        Place any custom initialiastion steps here.

        """

        tupleObservable = makeTupleDataObservableHandler(self.dbSessionCreator)

        self._loadedObjects.extend(
            makeAdminBackendHandlers(tupleObservable, self.dbSessionCreator))

        self._loadedObjects.append(tupleObservable)

        # session = self.dbSessionCreator()
        #
        # This will retrieve all the settings
        # from peek_plugin_livedb._private.storage.Setting import globalSetting
        # allSettings = globalSetting(session)
        # logger.debug(allSettings)
        #
        # This will retrieve the value of property1
        # from peek_plugin_livedb._private.storage.Setting import PROPERTY1
        # value1 = globalSetting(session, key=PROPERTY1)
        # logger.debug("value1 = %s" % value1)
        #
        # This will set property1
        # globalSetting(session, key=PROPERTY1, value="new value 1")
        # session.commit()
        #
        # session.close()

        mainController = MainController(
            dbSessionCreator=self.dbSessionCreator,
            tupleObservable=tupleObservable)

        self._loadedObjects.append(mainController)
        self._loadedObjects.append(makeTupleActionProcessorHandler(mainController))

        liveDbController = LiveDbController(self.dbSessionCreator)
        self._loadedObjects.append(liveDbController)

        liveDbImportController = LiveDbImportController(self.dbSessionCreator)
        self._loadedObjects.append(liveDbImportController)

        # Initialise the API object that will be shared with other plugins
        self._api = LiveDBApi(liveDbController=liveDbController,
                              liveDbImportController=liveDbImportController,
                              dbSessionCreator=self.dbSessionCreator,
                              dbEngine=self.dbEngine)
        self._loadedObjects.append(self._api)

        # noinspection PyTypeChecker
        liveDbImportController.setReadApi(self._api.readApi)

        logger.debug("Started")

    def stop(self):
        """ Stop

        This method is called by the platform to tell the peek app to shutdown and stop
        everything it's doing
        """
        # Shutdown and dereference all objects we constructed when we started
        while self._loadedObjects:
            self._loadedObjects.pop().shutdown()

        self._api = None

        logger.debug("Stopped")

    def unload(self):
        """Unload

        This method is called after stop is called, to unload any last resources
        before the PLUGIN is unlinked from the platform

        """
        logger.debug("Unloaded")

    @property
    def publishedServerApi(self) -> object:
        """ Published Server API
    
        :return  class that implements the API that can be used by other Plugins on this
        platform service.
        """
        return self._api

    ###### Implement PluginServerWorkerEntryHookABC

    @property
    def celeryApp(self) -> Celery:
        from peek_plugin_livedb._private.worker.CeleryApp import celeryApp
        return celeryApp



from peek_plugin_livedb._private.server.LiveDBReadApi import LiveDBReadApi
from peek_plugin_livedb._private.server.LiveDBWriteApi import LiveDBWriteApi
from peek_plugin_livedb._private.server.controller.LiveDbController import \
    LiveDbController
from peek_plugin_livedb._private.server.controller.LiveDbImportController import \
    LiveDbImportController
from peek_plugin_livedb.server.LiveDBApiABC import LiveDBApiABC
from peek_plugin_livedb.server.LiveDBReadApiABC import LiveDBReadApiABC
from peek_plugin_livedb.server.LiveDBWriteApiABC import LiveDBWriteApiABC


class LiveDBApi(LiveDBApiABC):
    def __init__(self, liveDbController: LiveDbController,
                 liveDbImportController: LiveDbImportController,
                 dbSessionCreator,
                 dbEngine):
        self._readApi = LiveDBReadApi(liveDbController=liveDbController,
                                      dbSessionCreator=dbSessionCreator,
                                      dbEngine=dbEngine)
        self._writeApi = LiveDBWriteApi(liveDbController=liveDbController,
                                        liveDbImportController=liveDbImportController,
                                        readApi=self._readApi,
                                        dbSessionCreator=dbSessionCreator,
                                        dbEngine=dbEngine)

    def shutdown(self):
        self._readApi.shutdown()
        self._writeApi.shutdown()

        self._readApi = None
        self._writeApi = None

    @property
    def writeApi(self) -> LiveDBWriteApiABC:
        return self._writeApi

    @property
    def readApi(self) -> LiveDBReadApiABC:
        return self._readApi



import logging

from txhttputil.util.DeferUtil import deferToThreadWrap
from typing import Union

from twisted.internet.defer import Deferred

from vortex.DeferUtil import deferToThreadWrapWithLogger
from vortex.Payload import Payload
from vortex.TupleSelector import TupleSelector
from vortex.handler.TupleDataObservableHandler import TuplesProviderABC

from peek_plugin_livedb._private.storage.StringIntTuple import StringIntTuple

logger = logging.getLogger(__name__)

class StringIntTupleProvider(TuplesProviderABC):
    def __init__(self, ormSessionCreator):
        self._ormSessionCreator = ormSessionCreator

    @deferToThreadWrapWithLogger(logger)
    def makeVortexMsg(self, filt: dict,
                      tupleSelector: TupleSelector) -> Union[Deferred, bytes]:
        # Potential filters can be placed here.
        # val1 = tupleSelector.selector["val1"]

        session = self._ormSessionCreator()
        try:
            tasks = (session.query(StringIntTuple)
                # Potentially filter the results
                # .filter(StringIntTuple.val1 == val1)
                .all()
            )

            # Create the vortex message
            return Payload(filt, tuples=tasks).makePayloadEnvelope().toVortexMsg()

        finally:
            session.close()



from typing import List, Dict

from sqlalchemy import select

from peek_plugin_base.storage.StorageUtil import makeCoreValuesSubqueryCondition, \
    makeOrmValuesSubqueryCondition
from peek_plugin_livedb._private.storage.LiveDbItem import LiveDbItem
from peek_plugin_livedb._private.storage.LiveDbModelSet import LiveDbModelSet, \
    getOrCreateLiveDbModelSet
from peek_plugin_livedb.tuples.LiveDbDisplayValueTuple import LiveDbDisplayValueTuple
from peek_plugin_livedb.tuples.LiveDbRawValueTuple import LiveDbRawValueTuple


class WorkerApi:
    """ Worker Api

    This class allows other classes to work with the LiveDB plugin on the
    worker service.

    """
    _FETCH_SIZE = 5000

    @classmethod
    def getLiveDbDisplayValues(cls,
                               ormSession,
                               modelSetName: str,
                               liveDbKeys: List[str]
                               ) -> List[LiveDbRawValueTuple]:
        """ Get Live DB Display Values

        Return an array of items representing the display values from the LiveDB.

        :param ormSession: The SQLAlchemy orm session from the calling code.
        :param modelSetName: The name of the model set to get the keys for
        :param liveDbKeys: An array of LiveDb Keys.

        :returns: An array of tuples.
        """
        if not liveDbKeys:
            return []

        liveDbModelSet = getOrCreateLiveDbModelSet(ormSession, modelSetName)

        liveDbKeys = set(liveDbKeys)  # Remove duplicates if any exist.
        qry = (
            ormSession.query(LiveDbItem)
                .filter(LiveDbItem.modelSetId == liveDbModelSet.id)
                .filter(makeOrmValuesSubqueryCondition(
                ormSession, LiveDbItem.key, list(liveDbKeys)
            ))
                .yield_per(cls._FETCH_SIZE)
        )

        results = []

        for item in qry:
            results.append(
                LiveDbDisplayValueTuple(key=item.key,
                                        displayValue=item.displayValue,
                                        rawValue=item.rawValue,
                                        dataType=item.dataType)
            )

        return results

    @classmethod
    def getLiveDbKeyDatatypeDict(cls, ormSession,
                                 modelSetName: str,
                                 liveDbKeys: List[str]) -> Dict[str, int]:
        """ Get Live DB Display Values

        Return an array of items representing the display values from the LiveDB.

        :param ormSession: The SQLAlchemy orm session from the calling code.
        :param modelSetName: The name of the model set to get the keys for
        :param liveDbKeys: An array of LiveDb Keys.

        :returns: An array of tuples.
        """
        liveDbTable = LiveDbItem.__table__
        modelTable = LiveDbModelSet.__table__

        if not liveDbKeys:
            return {}

        liveDbKeys = list(set(liveDbKeys))  # Remove duplicates if any exist.
        stmt = (select([liveDbTable.c.key, liveDbTable.c.dataType])
                .select_from(liveDbTable
                             .join(modelTable,
                                   liveDbTable.c.modelSetId == modelTable.c.id))
                .where(modelTable.c.name == modelSetName)
                .where(makeCoreValuesSubqueryCondition(
                    ormSession.bind, liveDbTable.c.key, liveDbKeys
                ))
        )
        resultSet = ormSession.execute(stmt)
        return dict(resultSet.fetchall())
